#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# =============================================================================
# ___  ___                       _____           _______
# |  \/  |                      |_   _|         | | ___ \
# | .  . |_   _ ___  ___  ___     | | ___   ___ | | |_/ / _____  __
# | |\/| | | | / __|/ _ \/ _ \    | |/ _ \ / _ \| | ___ \/ _ \ \/ /
# | |  | | |_| \__ \  __/ (_) |   | | (_) | (_) | | |_/ / (_) >  <
# \_|  |_/\__,_|___/\___|\___/    \_/\___/ \___/|_\____/ \___/_/\_\
#
# @author:  Nicolas Karasiak
# @site:    www.karasiak.net
# @git:     www.github.com/nkarasiak/MuseoToolBox
# =============================================================================
import os
import numpy as np
from osgeo import ogr
from .. import processing


class distanceCV:
    def __init__(
            self,
            X,
            y,
            distance_matrix,
            distance_thresold=False,
            valid_size=1,
            n_repeats=False,
            n_splits=False,
            verbose=False,
            random_state=False,
            groups=None,
            distance_label=False,
            LOO_same_size=False):
        """Compute train/validation array with Spatial distance analysis.
        Object stops when less effective class number is reached (45 loops if your least class contains 45 ROI).
        Parameters
        ----------
        X : array-like
            Matrix of values. As many row as the Y array.
        Y : array-like
            contain class for each ROI. Same effective as distance_matrix.
        distance_matrix : array
            Matrix distance
        distance_thresold : int, float or False, optional (default=False).
            Distance(same unit of your distance_matrix).
            If False, will split spatially the dataset.
        n_repeats : int or False, optional (default=False).
            False : as loop as min effective class
        n_splits : int or False, optional (default=False).
            Number of split per CV. Default is the size of the smallest class.
        valid_size : False or float value.
            If float, value from 0 to 1 (percent).
        groups : array
            contain class (like Y), e.g. able to do a SLOO per Stand if you put your stand number here.
        stat : str.
            Path where to save csv stat (label, n trains, and mean distance between training per class).
        LOO_same_size : bool, optional (default=False).
            Use the same training size as SLOO but use samples without taking into account the distance.

        Returns
        -------
        train : array
            List of Y selected ROI for train
        validation : array
            List of Y selected ROI for validation
        """
        self.name = 'SLOO'

        self.distance_matrix = distance_matrix
        self.distance_thresold = distance_thresold
        self.y = y
        self.iterPos = 0

        self.y_labels = np.unique(y, return_counts=True)[0]

        self.distance_label = distance_label

        self.valid_size = valid_size
        self.LOO_same_size = LOO_same_size
        self.groups = groups

        self.verbose = verbose

        self.random_state = random_state

        self.mask = np.ones(np.asarray(self.y).shape, dtype=bool)

        if self.groups is None:
            self.minEffectiveClass = min(
                [len(self.y[self.y == i]) for i in np.unique(self.y)])
        else:
            self.minEffectiveClass = min(
                [len(np.unique(groups[np.where(y == i)[0]])) for i in np.unique(self.y)])

        if n_splits:
            self.minEffectiveClass = n_splits

        if n_repeats and self.valid_size >= 1:
            self.n_repeats = self.minEffectiveClass * n_repeats

        elif self.valid_size >= 1:
            # TODO : run self.__next__() to get real n_repeats as it depends on distance
            # but if running self.__next__() here, iterator will be empty
            # after.
            self.n_repeats = self.minEffectiveClass
        elif isinstance(self.valid_size, float):
            self.n_repeats = int(1 / (self.valid_size))
        else:
            self.n_repeats = self.minEffectiveClass

    def __iter__(self):
        return self

    # python 3 compatibility

    def __next__(self):
        return self.next()

    def next(self):
        emptyTrain = False

        if self.iterPos < self.n_repeats:
            self.nTries = 0
            completeTrain = False
            while completeTrain is False:
                if self.nTries < 100:
                    emptyTrain = False
                    if self.verbose:
                        print(53 * '=')
                    validation, train = np.array([[], []], dtype=np.int64)
                    for C in self.y_labels:
                        # Y is True, where C is the unique class
                        CT = np.where(self.y == C)[0]

                        currentCT = np.logical_and(self.y == C, self.mask == 1)

                        if np.where(currentCT)[
                                0].shape[0] == 0:  # means no more ROI
                            if self.verbose > 1:
                                print(
                                    str(C) + ' has no more valid pixel.\nResetting label mask.')
                            self.mask[self.y == C] = 1
                            currentCT = np.logical_and(
                                self.y == C, self.mask == 1)

                        if np.where(currentCT)[
                                0].shape[0] == 0:  # means no more ROI
                            if self.verbose > 1:
                                print(
                                    str(C) + ' has no more valid pixel. Reusing samples.')
                            self.mask[self.y == C] = 1
                            currentCT = np.logical_and(
                                self.y == C, self.mask == 1)

                        np.random.seed(self.random_state)
                        self.ROI = np.random.permutation(
                            np.where(currentCT)[0])[0]
                        # When doing Leave-One-Out per subgroup
                        if self.groups is not None:
                            if self.verbose > 1:
                                print('ROI stand is ' +
                                      str(self.groups[self.ROI]) +
                                      ' for label ' +
                                      str(C))
                            standPos = np.argwhere(
                                self.groups[self.ROI] == self.distance_label)[0][0]
                            distanceROI = (self.distance_matrix[standPos, :])
                            tmpValid = np.where(
                                self.groups == self.groups[self.ROI])[0].astype(np.int64)

                            tmpTrainGroup = np.unique(
                                self.distance_label[np.where(distanceROI >= self.distance_thresold)[0]])
                            tmpTrainGroup = tmpTrainGroup[np.isin(
                                tmpTrainGroup, self.groups[CT])]
                            tmpTrain = np.where(np.in1d(self.groups, tmpTrainGroup))[
                                0].flatten()

                            if tmpTrain.shape[0] == 0:
                                emptyTrain = True
                        # When doing Leave-One-Out
                        else:

                            # get line of distance for specific ROI
                            distanceROI = (
                                self.distance_matrix[self.ROI, :])[CT]
                            if self.valid_size is False:
                                tmpValid = np.array(
                                    [self.ROI], dtype=np.int64)
                                tmpTrain = CT[distanceROI >
                                              self.distance_thresold]
                                if self.LOO_same_size is True:
                                    np.random.seed(self.random_state)
                                    tmpTrain = np.random.permutation(
                                        CT)[:len(tmpTrain)]

                            else:
                                if self.valid_size >= 1:
                                    nToCut = self.valid_size
                                else:
                                    nToCut = int(self.valid_size * len(CT))

                                if self.distance_thresold is False:
                                    distanceToCut = np.sort(distanceROI)[
                                        :nToCut][-1]
                                    tmpValid = CT[distanceROI <=
                                                  distanceToCut]

                                    tmpTrain = CT[distanceROI >
                                                  distanceToCut]

                                else:
                                    tmpValid = np.asarray([self.ROI])

                                    tmpTrain = CT[distanceROI >
                                                  self.distance_thresold]

                                if self.LOO_same_size is True:
                                    np.random.seed(self.random_state)

                            if tmpTrain.shape[0] == 0:
                                emptyTrain = True

                        del CT
                        if emptyTrain is True:
                            if self.verbose:
                                print(
                                    'No training for class ' +
                                    str(C) +
                                    '. Doing a new fold.')

                            self.mask[tmpValid] = 0

                        else:
                            #
                            validation = np.concatenate((validation, tmpValid))
                            train = np.concatenate((train, tmpTrain))

                    if self.verbose:
                        print('Validation samples : ' + str(len(validation)))
                        print('Training samples : ' + str(len(train)))

                    # Mask selected validation
                    self.random_state += 1
                    if emptyTrain is True:
                        completeTrain = False
                        self.nTries += 1

                    else:
                        self.iterPos += 1
                        self.mask[validation] = 0
                        return train, validation
                else:
                    raise ValueError(
                        'Error : Not enough samples using this distance/valid_size.')

        else:
            raise StopIteration()


class randomPerClass:
    """
    Random array according to Y.

    Parameters
    ----------
    X : None
    Y : arr, default None.
        Label for each feature.
    groups : arr, default None.
        Group for each feature. For sklearn compatibility.
    valid_size : False or int
        1 to do a Leave-One-Out.
    train_size : float (<1.0) or int (>1).
        Percentage to keep for training or integer.
    n_splits : False or int.
        If False, will be the number of samples of the smallest class.
    random_state : int.
        random_state for numpy.
    verbose : boolean or int.

    Returns
    -------
    train,validation : array of indices
    """

    def __init__(self, X=None, y=None, groups=None,
                 valid_size=0.5, train_size=None, n_repeats=False,
                 random_state=None, verbose=False):
        self.y = y
        self.valid_size = valid_size
        self.train_size = 1 - self.valid_size

        smallestClass = np.min(np.unique(y, return_counts=True)[1])

        if n_repeats is False:
            if self.valid_size >= 1:
                self.n_repeats = smallestClass
            else:
                self.n_repeats = int(1 / self.valid_size)
        else:
            self.n_repeats = n_repeats

        if self.valid_size < 1:
            test_n_splits = int(valid_size * smallestClass)
            if test_n_splits == 0:
                raise ValueError('Valid size is too small')

        if groups is not None and verbose:
            print("Received groups value, but randomPerClass don't use it")

        self.n_splits = self.n_repeats

        self.random_state = random_state
        self.iterPos = 1
        self.mask = np.ones(np.asarray(self.y).shape, dtype=bool)

    def __iter__(self):
        return self

    # python 3 compatibility
    def __next__(self):
        return self.next()

    def next(self):
        if self.iterPos < self.n_repeats + 1:

            train, valid = [np.asarray(
                [], dtype=int), np.asarray([], dtype=int)]
            for C in np.unique(self.y):
                Cpos = np.where(self.y == C)[0]
                np.random.seed(self.random_state)

                if self.valid_size < 1:  # means in percent
                    nToKeep = int(self.valid_size * len(Cpos))
                else:
                    nToKeep = self.valid_size

                unMask = np.logical_and(self.y == C, self.mask == 1)
                tmpValid = np.random.permutation(
                    np.where(unMask == 1)[0])[:nToKeep]

                TF = np.in1d(Cpos, tmpValid, invert=True)
                tmpTrain = Cpos[TF]

                train = np.concatenate((train, tmpTrain))
                valid = np.concatenate((valid, tmpValid))

                self.mask[tmpValid] = 0

                unMask = np.logical_and(self.y == C, self.mask == 1)
                if self.valid_size < 1:
                    if np.where(unMask == 1)[0].shape[0] < int(
                            self.valid_size * len(Cpos)):
                        self.mask[np.where(self.y == C)[0]] = 1
                else:
                    if np.where(unMask == 1)[0].shape[0] < self.valid_size:
                        self.mask[np.where(self.y == C)[0]] = 1

            self.random_state += 1
            self.iterPos += 1

            return train, valid
        else:
            raise StopIteration()


class groupCV:
    def __init__(self, X=None, y=None, groups=None, n_repeats=False,
                 valid_size=1, random_state=False, verbose=False):
        """Compute train/validation per group.

        Parameters
        ----------

        X : None
            For sklearn compatiblity
        Y : array-like
            contains class for each ROI.
        groups : array-like
            contains goup number for each ROI (same size of Y).
        valid_size : int (1) or float (0.01 to 0.99)
            If 1 Leave-One-Group Out.
        n_splits : False or int
            if False, n_splits is the minimum stand number of all species.
        SLOO :  Bool
            True  or False. If SLOO, keep only one Y per validation stand.
        """
        self.verbose = verbose
        self.y = y.flatten()
        self.uniqueY = np.unique(self.y)
        self.groups = groups

        self.valid_size = valid_size
        self.iterPos = 1

        self.random_state = random_state

        smallestGroup = []
        for i in np.unique(self.y):
            standNumber = np.unique(
                np.array(groups)[
                    np.where(
                        np.array(self.y).flatten() == i)])

            smallestGroup.append(standNumber.shape[0])
        smallestGroup = np.min(smallestGroup)

        if n_repeats:
            self.n_repeats = n_repeats
        else:
            self.n_repeats = smallestGroup
            if self.n_repeats == 1:
                raise Exception(
                    'You need to have at least two subgroups per label')

        test_n_splits = np.amax(
            (int(valid_size * smallestGroup), int((1 - valid_size) * smallestGroup)))
        if test_n_splits == 0:
            raise ValueError('Valid size is too small')

        self.mask = np.ones(np.asarray(groups).shape, dtype=bool)

        self.n_splits = self.n_repeats

    def __iter__(self):
        return self

    # python 3 compatibility
    def __next__(self):
        return self.next()

    def next(self):
        if self.iterPos < self.n_repeats + 1:
            if self.verbose:
                print(53 * '=')
            train = np.array([], dtype=int)
            validation = np.array([], dtype=int)
            for C in self.uniqueY:
                Ycurrent = np.where(np.array(self.y) == C)[0]
                Ystands = np.array(self.groups)[Ycurrent]

                nYstand = len(np.unique(self.groups[self.y == C]))
                Ystand = self.groups[np.logical_and(
                    self.mask == 1, self.y == C)]
                if self.valid_size >= 1:
                    nToKeep = self.valid_size
                else:
                    nToKeep = nYstand * self.valid_size
                    nToKeep += (nToKeep < 1)
                    nToKeep = int(nToKeep)

                if np.unique(Ystand).shape[0] < nToKeep:
                    # reset mask because not enough group
                    self.mask[Ycurrent] = 1
                    Ystand = self.groups[self.y == C]

                np.random.seed(self.random_state)
                selectedStand = np.random.permutation(
                    np.unique(Ystand))[:nToKeep]

                if self.verbose:
                    print('For class {}, subgroup {}'.format(C, selectedStand))

                YinSelectedStandt = np.in1d(Ystands, selectedStand)
                tmpValid = Ycurrent[YinSelectedStandt]
                validation = np.concatenate(
                    (validation, tmpValid))

                YnotInSelectedStandt = np.invert(YinSelectedStandt)
                tmpTrain = Ycurrent[YnotInSelectedStandt]
                train = np.concatenate(
                    (train, tmpTrain))

                self.mask[tmpValid] = 0
            self.random_state += 1
            self.iterPos += 1
            return train, validation
        else:
            raise StopIteration()


class _cv_manager:
    def __init__(self, cv_type, verbose=False, **params):
        """
        Manage cross-validation methods to generate the duo valid/train samples.

        """
        self.cv_type = cv_type
        self.verbose = verbose

        self.__extensions = ['sqlite', 'shp', 'netcdf', 'gpx', 'gpkg']
        self.__driversName = [
            'SQLITE',
            'ESRI Shapefile',
            'netCDF',
            'GPX',
            'GPKG']

        self.__alertMessage = 'It seems you already generated the cross validation. \n Please use reinitialize function if you want to regenerate the cross validation. \n But check if you defined a random_state number in your samplingMethods function to have the same output.'
        self.__alreadyRead = False

        self.params = params

        # create unique random state if no random_state
        if self.params['random_state'] is False:
            import time
            self.params['random_state'] = int(time.time())

    def reinitialize(self):
        _cv_manager.__init__(
            self, self.cv_type, self.verbose, **self.params)

    def get_supported_extensions(self):
        print('Museo ToolBox supported extensions for writing vector files are : ')
        for idx, ext in enumerate(self.__extensions):
            print(3 * ' ' + '- ' + self.__driversName[idx] + ' : ' + ext)

    def get_n_splits(self, X=None, y=None, groups=None):
        """
        Returns the number of splitting iterations in the cross-validator.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features), optional
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like, of length n_samples
            The target variable for supervised learning problems.
        groups : array-like, with shape (n_samples,), optional
            Subgroup labels for the samples used while splitting the dataset into
            train/test set.

        Returns
        -------
        n_splits : int
            The number of splits.
        """

        if X is not None:
            X = np.empty(y.reshape(-1, 1).shape, dtype=np.int16)

        if self.cv_type.__name__ == 'distanceCV':
            # TODO : Find a better way to get n_splits for distanceCV
            # As distance may differ from real n_splits, hard to not run the
            # whole thing
            n_splits = 0
            for tr, vl in self.cv_type(
                    X=X, y=y, groups=groups, verbose=self.verbose, **self.params):
                n_splits += 1

        else:
            n_splits = self.cv_type(
                X=X, y=y, groups=groups, verbose=self.verbose, **self.params).n_splits

        if n_splits == 0:
            raise ValueError(
                'Sorry but due to your dataset or your distance thresold, no cross-validation has been generated.')

        return n_splits

    def split(self, X, y, groups=None):
        """
        Split the vector/array according to y and groups.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features), optional
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like, of length n_samples
            The target variable for supervised learning problems.
        groups : array-like, with shape (n_samples,), optional
            Subgroup labels for the samples used while splitting the dataset into
            train/test set.

        Returns
        --------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.

        """
        y = y.reshape(-1, 1)
        if self.__alreadyRead:
            self.reinitialize()
        self.__alreadyRead = True

        return self.cv_type(
            X=X, y=y, groups=groups, verbose=self.verbose, **self.params)

    def save_to_vector(self, vector, field, group=None, out_vector=None):
        """
        Save to vector files each fold from the cross-validation.

        Parameters
        -------------
        vector : str.
            Path where the vector is stored.
        field : str.
            Name of the field containing the label.
        group : str, or None.
            Name of the field containing the group/subgroup (or None
        out_vector : str.

            Path and filename to save the different results.

        Returns
        ----------
        list_of_output : list
            List containing the number of folds * 2.

            train + validation for each fold.

        """
        src = ogr.Open(vector)
        srcLyr = src.GetLayerByIndex()
        self.wkbType = srcLyr.GetGeomType()
        if self.wkbType != 1:
            print("""Warning : This function generates vector files according to your vector.
        The number of features may differ from the number of pixels used in classification.
        If you want to save every ROI pixels in the vector, please use processing.sample_extraction before.""")
        del src, srcLyr

        fileName, self.__ext = os.path.splitext(out_vector)

        if self.__ext[1:] not in self.__extensions:
            print(
                'Your extension {} is not recognized as a valid extension for saving shape.'.format(self.__ext))
            self.get_supported_extensions()
            raise Exception(
                'We recommend you to use gpkg or sqlite extension.')

        if group is None:
            groups = None
            y, fts, srs = processing.read_vector_values(
                vector, field, get_features=True, verbose=self.verbose)
        else:
            y, groups, fts, srs = processing.read_vector_values(
                vector, field, group, get_features=True, verbose=self.verbose)

        if self.__alreadyRead:
            self.reinitialize()
        listOutput = []
        self.cv = []
        for idx, trvl in enumerate(self.split(None, y, groups)):
            if trvl is not None:
                self.cv.append([trvl[0], trvl[1]])
                trFeat = [fts[int(i)] for i in trvl[0]]
                vlFeat = [fts[int(i)] for i in trvl[1]]
                tr = fileName + '_train_' + str(idx) + self.__ext
                vl = fileName + '_valid_' + str(idx) + self.__ext
                self.__save_to_vector(trFeat, srs, tr)
                self.__save_to_vector(vlFeat, srs, vl)
                listOutput.append([tr, vl])
        self.__alreadyRead = True
        return listOutput

    def __save_to_vector(self, array, srs, outShapeFile):
        # Parse a delimited text file of volcano data and create a shapefile
        # use a dictionary reader so we can access by field name
        # set up the shapefile driver

        driverIdx = [x for x, i in enumerate(
            self.__extensions) if i == self.__ext[1:]][0]
        outDriver = ogr.GetDriverByName(self.__driversName[driverIdx])

        if os.path.exists(outShapeFile):
            # Remove output shapefile if it already exists
            outDriver.DeleteDataSource(outShapeFile)

        # create the data source
        ds = outDriver.CreateDataSource(outShapeFile)

        # create the spatial reference, WGS84
        lyrout = ds.CreateLayer(
            'cv',
            srs=srs,
            geom_type=self.wkbType)
        fields = [
            array[1].GetFieldDefnRef(i).GetName() for i in range(
                array[1].GetFieldCount())]

        if self.__ext[1:] != 'shp':
            isShp = False
            lyrout.StartTransaction()
        else:
            isShp = True

        for i, f in enumerate(fields):
            field_name = ogr.FieldDefn(
                f, array[1].GetFieldDefnRef(i).GetType())
            if isShp:
                field_name.SetWidth(array[1].GetFieldDefnRef(i).GetWidth())
            lyrout.CreateField(field_name)

        for k in array:
            lyrout.CreateFeature(k)
        k, i, f = None, None, None

        if not isShp:
            lyrout.CommitTransaction()
        # Save and close the data source
        ds = None



# -*- coding: utf-8 -*-
import unittest

import numpy as np
from museotoolbox import ai
from museotoolbox.datasets import load_historical_data
from museotoolbox.processing import image_mask_from_vector
from osgeo import gdal

import os
import shutil

from sklearn.ensemble import RandomForestClassifier

raster,vector = load_historical_data(low_res=True)
X,y,g = load_historical_data(return_X_y_g=True,low_res=True)
param_grid = dict(n_estimators=[1,10])
classifier = RandomForestClassifier()
image_mask_from_vector(vector,raster,'/tmp/mask.tif')

class TestStats(unittest.TestCase):
    def test_superLearner(self):
        
        n_cv = 2
        for tf in [True,False]:
            verbose = tf+1
            model = ai.SuperLearner(classifier,param_grid=param_grid,n_jobs=1,verbose=verbose)
            model.fit(X,y,group=g,standardize=tf,cv=n_cv)
            assert(model.predict_array(X).shape == y.shape)
            len(model.CV) == n_cv
            assert(np.all(model.group == g))
            
            model.predict_image(raster,'/tmp/SuperLearner/class.tif',confidence_per_class='/tmp/SuperLearner/confclass.tif',higher_confidence='/tmp/SuperLearner/higherconf.tif')
            assert(model._is_standardized == tf)
        
        # test masked return if X is totally masked
        X_masked = np.ma.copy(X)
        X_masked.mask=True
        X_masked_return = model._convert_array(X_masked)
        assert(np.ma.is_masked(X_masked_return))
        

        
    def test_superLearn_pred(self):
        model = ai.SuperLearner(classifier,param_grid=param_grid,n_jobs=1,verbose=0)
        model.customize_array(np.mean,axis=1)
        model.fit(X,y,group=g,standardize=True,cv=2)
#        #    
        assert(model._array_is_customized == True)
        assert(model._array_is_customized)
        assert(model.xFunction)
        assert(np.all(model.standardize_array(X) != X))
        model.predict_image(raster,out_image='/tmp/SuperLearner/class.tif',in_image_mask='/tmp/mask.tif',confidence_per_class='/tmp/SuperLearner/confclass.tif',higher_confidence='/tmp/SuperLearner/higherconf.tif')
        assert(gdal.Open('/tmp/SuperLearner/class.tif').RasterCount == 1)
        assert(gdal.Open('/tmp/SuperLearner/higherconf.tif').RasterCount == 1)
        assert(gdal.Open('/tmp/SuperLearner/confclass.tif').RasterCount == len(np.unique(y)))
        cms = model.get_stats_from_cv()
        
        assert(len(cms) == 2)
        model.save_cm_from_cv('/tmp/empty/',prefix='coco',header=False)
        shutil.rmtree('/tmp/empty/')
        model.save_model('/tmp/SuperLearner/model')
        assert(os.path.exists('/tmp/SuperLearner/model.npz'))
        model.load_model('/tmp/SuperLearner/model')
        shutil.rmtree('/tmp/SuperLearner/')

        with self.assertRaises(ValueError):
            model.fit(X,y,cv=False)
        
    def test_sequential(self):
        
        sfs = ai.SequentialFeatureSelection(classifier,param_grid)
        sfs.fit(X,y,cv=2)
        sfs.predict(X,idx=0)
        assert(not np.all(sfs.predict(X,idx=0) == sfs.predict(X,idx=1)))
        sfs.predict_best_combination(raster,'/tmp/class.tif')
        sfs.predict_images(raster,'/tmp/class')
        assert(sfs.get_best_model())
        assert(sfs.transform(X,idx='best').shape[1] == sfs.best_idx_+1)

        n_comp = 2
        max_features = 2
        sfs = ai.SequentialFeatureSelection(classifier,param_grid,n_comp=n_comp,path_to_save_models='/tmp/sfs_models/',verbose=1)
        def double_columns(x):
            return np.hstack((x,x))
        sfs.customize_array(double_columns)
        sfs.fit(X,y,max_features=max_features,standardize=True,cv=2)
        sfs.fit(X,y,max_features=max_features,standardize=True,cv=2) # to reload from path
        assert(sfs.transform(X,idx=1).shape[1] == 2*n_comp)
        assert(sfs.transform(X,idx=0).shape[1] == n_comp)
        assert(sfs.X.shape[1] == X.shape[1]*2)
        assert(len(sfs.best_features_) == 2)
        sfs.predict_images(raster,'/tmp/sfs_models/')
        sfs.predict_best_combination(raster,'/tmp/sfs_models/best.tif')
        assert(sfs.get_best_model().X.shape[1] == n_comp*(sfs.best_idx_+1) )
        sfs.predict(X,0)
        shutil.rmtree('/tmp/sfs_models/')
        
if __name__ == "__main__":
    unittest.main()
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# =============================================================================
# ___  ___                       _____           _______
# |  \/  |                      |_   _|         | | ___ \
# | .  . |_   _ ___  ___  ___     | | ___   ___ | | |_/ / _____  __
# | |\/| | | | / __|/ _ \/ _ \    | |/ _ \ / _ \| | ___ \/ _ \ \/ /
# | |  | | |_| \__ \  __/ (_) |   | | (_) | (_) | | |_/ / (_) >  <
# \_|  |_/\__,_|___/\___|\___/    \_/\___/ \___/|_\____/ \___/_/\_\
#
# @author:  Nicolas Karasiak
# @site:    www.karasiak.net
# @git:     www.github.com/nkarasiak/MuseoToolBox
# =============================================================================
import re

__version__ = re.search(
    r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',  # It excludes inline comment too
    open('museotoolbox/__init__.py').read()).group(1)

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name='museotoolbox',
    version=__version__,
    description='Raster and vector tools for Remote Sensing and Classification',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/nkarasiak/MuseoToolBox',
    author='Nicolas Karasiak',
    author_email='karasiak.nicolas@gmail.com',
    license='GPLv3',
    install_requires=["numpy","scipy","matplotlib","scikit-learn",'joblib'],
    packages=setuptools.find_packages(),
    classifiers=[
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: GIS",
            "Programming Language :: Python :: 3",
            "Intended Audience :: Science/Research"],
    zip_safe=False,
    package_data={
      'museotoolbox': ['datasets/_historicalmap/map_compress.tif','datasets/_historicalmap/map_lowres.tif','datasets/_historicalmap/train.gpkg','datasets/_historicalmap/train_centroid.gpkg']
   }
 )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# =============================================================================
# ___  ___                       _____           _______
# |  \/  |                      |_   _|         | | ___ \
# | .  . |_   _ ___  ___  ___     | | ___   ___ | | |_/ / _____  __
# | |\/| | | | / __|/ _ \/ _ \    | |/ _ \ / _ \| | ___ \/ _ \ \/ /
# | |  | | |_| \__ \  __/ (_) |   | | (_) | (_) | | |_/ / (_) >  <
# \_|  |_/\__,_|___/\___|\___/    \_/\___/ \___/|_\____/ \___/_/\_\
#
# @author:  Nicolas Karasiak
# @site:    www.karasiak.net
# @git:     www.github.com/nkarasiak/MuseoToolBox
# =============================================================================
"""
The :mod:`museotoolbox.charts` module gathers plotting functions.
"""

from matplotlib import pyplot as plt
from matplotlib import gridspec
import numpy as np
import itertools

np.seterr(divide='ignore', invalid='ignore')

# for numpy version < 1.17


def _nan_to_num(array, nan=0):
    return np.where(np.isnan(array), nan, array)


class PlotConfusionMatrix:
    """
    Plot a confusion matrix with imshow of pyplot.
    Customize color (e.g. diagonal color), add subplots with F1 or Producer/User accuracy.

    Examples
    --------
    >>> plot = mtb.charts.plotConfusionMatrix([[5,6],[1,8]])
    >>> plot.add_text()
    >>> plot.add_f1()
    """

    def __init__(self, cm, cmap=plt.cm.Greens,
                 left=None, right=None, **kwargs):
        self.cm = np.array(cm)
        self.cm_ = np.copy(cm)
        self.axes = []
        self.gs = gridspec.GridSpec(
            2, 2, width_ratios=[
                self.cm.shape[1], 1], height_ratios=[
                self.cm.shape[0], 1])

        self.gs.update(
            bottom=0,
            top=1,
            wspace=0,
            hspace=0.7 / self.cm.shape[0], right=right, left=left)

        self.ax = plt.subplot(self.gs[0, 0])  # place it where it should be.
        self.vmin = np.amin(self.cm)
        self.vmax = np.amax(self.cm)

        self.xlabelsPos = 'bottom'
        self.xrotation = 0
        self.yrotation = 0
        self.font_size = False

        self.cmap = cmap
        self.diag_color = cmap
        self.ax.set_yticks(range(self.cm.shape[0]))

        self.fig = plt.figure(1)

        self.ax.imshow(
            cm,
            interpolation='nearest',
            aspect='equal',
            cmap=self.cmap,
            vmin=self.vmin,
            vmax=self.vmax,
            **kwargs)

        self.kwargs = kwargs
        self.subplot = False
        self.axes.append(self.ax)

    def add_text(self, thresold=False, font_size=12, alpha=1, alpha_zero=1):
        """
        Add value of each case on the matrix image.

        Parameters
        ----------
        thresold : False or integer.
        alpha : float, default 1.
            Value from 0 to 1.
        alpha_zero : float, default 1.
            Value alpha for 0 values, from 0 to 1.

        Examples
        --------
        >>> plot.add_text(alpha_zero=0.5)
        """
        self.font_size = font_size
        if thresold is False:
            thresold = int(np.amax(self.cm) / 2)
        for i, j in itertools.product(
                range(self.cm.shape[0]), range(self.cm.shape[1])):
            if not np.ma.is_masked(self.cm[i, j]):
                # print(cm[i,j])
                self.ax.text(j, i, str(self.cm[i, j]),
                             horizontalalignment="center",
                             color="white" if self.cm[i, j] > thresold else 'black', fontsize=font_size, va='center', alpha=alpha_zero if self.cm[i, j] == 0 else alpha)
            else:
                #                print(self.cm2[i, j])
                self.ax.text(j, i, str(self.cm2[i, j]),
                             horizontalalignment="center",
                             color="white" if self.cm2[i, j] > thresold else "black", va='center', fontsize=font_size, )

    def add_x_labels(self, labels=None, rotation=90, position='top'):
        """
        Add labels for X.

        Parameters
        ----------
        labels : None
            If labels, best with same len as the X shape.
        rotation : int, default 90.
            Int, 45 or 90 is best.
        position : str, default 'top'.
            'top' or 'bottom'.

        Examples
        --------
        >>> plot.add_text(labels=['Tofu','Houmous'],alpha_zero=0.5,rotation=45)
        """
        self.xrotation = rotation
        self.xlabels = labels
        self.xlabelsPos = position
        if self.xlabelsPos == 'top':
            self.ax.xaxis.tick_top()
            self.ax.xaxis.set_ticks_position('top')  # THIS IS THE ONLY CHANGE

        self.ax.set_xticklabels(
            ['F1'],
            horizontalalignment='left',
            rotation=rotation, fontsize=self.font_size)

        self.ax.set_xticks(np.arange(self.cm.shape[1]))
        if rotation != 90:
            ha = 'left'
        else:
            ha = 'center'
        self.ax.set_xticklabels(
            self.xlabels,
            rotation=rotation,
            ha=ha,
            fontsize=self.font_size)

    def add_mean(self, xLabel='', yLabel='', hide_ticks=False,
                 thresold=50, vmin=0, vmax=100):
        """
        Add Mean for both axis.

        Parameters
        ----------
        xLabel : str
            The label for X (i.e. 'All species')
        yLabel : str
            The label for Y (i.e. 'All years')
        thresold : int, default 50.
        vmin : int.
            Minimum value for colormap.
        vmax :
            Maximum value for colormap.

        Examples
        --------
        >>> plot.add_mean(xLabel='all species',yLabel='all years')
        """
        if self.subplot is not False:
            raise Warning(
                'You can\'t add two subplots. You already had ' + str(self.subplot))
        else:
            self.subplot = 'Mean'

        self.ax1v = plt.subplot(self.gs[0, 1])
        self.ax1h = plt.subplot(self.gs[1, 0])

        valV = np.mean(self.cm_, axis=1).reshape(-1, 1).astype(int)

        valH = np.mean(self.cm_, axis=0).reshape(1, -1).astype(int)

        self.ax1v.imshow(
            valV,
            cmap=self.diag_color,
            interpolation='nearest',
            aspect='equal',
            vmin=vmin,
            vmax=vmax)
        self.ax1h.imshow(
            valH,
            cmap=self.diag_color,
            interpolation='nearest',
            aspect='equal',
            vmin=vmin,
            vmax=vmax)

        if hide_ticks:
            self.ax1v.set_yticks([])
        else:
            self.ax1v.set_yticks(np.arange(self.cm_.shape[0]))
        self.ax1v.set_xticks([])

        self.ax1h.set_yticks([0])
        self.ax1h.set_xticks([])

        for i in range(self.cm.shape[0]):

            iVal = np.int(np.mean(self.cm_, axis=1)[i])

            self.ax1v.text(0, i, iVal, color="white" if iVal >
                           thresold else 'black', ha='center', va='center', fontsize=self.font_size)

        self.ax1v.set_yticklabels([])
        for j in range(self.cm.shape[1]):
            jVal = np.int(np.mean(self.cm_, axis=0)[j])

            self.ax1h.text(j, 0, jVal, color="white" if jVal >
                           thresold else 'black', ha='center', va='center', fontsize=self.font_size)

        self.ax1h.set_yticklabels(
            [yLabel],
            rotation=self.yrotation,
            ha='right',
            va='center', fontsize=self.font_size)

        self.ax1v.xaxis.set_ticks_position('top')  # THIS IS THE ONLY CHANGE
        self.ax1v.set_xticks([0])
        if self.xrotation < 60:
            ha = 'left'
        else:
            ha = 'center'
        self.ax1v.set_xticklabels(
            [xLabel],
            horizontalalignment='left',
            rotation=self.xrotation,
            ha=ha,
            fontsize=self.font_size)
        self.axes.append([self.ax1v, self.ax1h])

    def add_y_labels(self, labels=None, rotation=0):
        """
        Add labels for Y.

        Parameters
        ----------
        labels : None
            If labels, best with same len as the X shape.
        rotation : int, default 90.
            Int, 45 or 90 is best.

        Examples
        --------
        >>> plot.add_y_labels(labels=['Fried','Raw'])
        """
        self.yrotation = rotation
        self.ylabels = labels
        self.ax.set_yticklabels(
            self.ylabels,
            rotation=rotation,
            horizontalalignment='right',
            fontsize=self.font_size)

    def add_f1(self):
        """
        Add F1 subplot.

        Examples
        --------
        >>> plot.add_f1()
        """
        if self.subplot is not False:
            raise Warning(
                'You can\'t add two subplots. You already had ' + str(self.subplot))
        if self.cm.shape[0] != self.cm.shape[1]:
            raise Warning('Number of lines and columns must be equal')

        self.subplot = 'F1'
        self.ax1v = plt.subplot(self.gs[0, 1])

        verticalPlot = []

        for label in range(self.cm.shape[0]):
            TP = self.cm_[label, label]
            #TN = np.sum(sp.diag(currentCsv))-currentCsv[label,label]
            FN = np.nansum(self.cm_[:, label]) - TP
            FP = np.nansum(self.cm_[label, :]) - TP

            verticalPlot.append(2 * TP / (2 * TP + FP + FN) * 100)

        verticalPlot = np.asarray(verticalPlot).reshape(-1, 1)
        self.ax1v.imshow(
            verticalPlot,
            cmap=self.diag_color,
            interpolation='nearest',
            aspect='equal',
            vmin=0,
            vmax=100)

        if self.xlabelsPos == 'top':
            self.ax1v.xaxis.tick_top()
            self.ax1v.xaxis.set_ticks_position(
                'top')  # THIS IS THE ONLY CHANGE
            self.ax1v.set_xticks([0])
            self.ax1v.set_xticklabels(
                ['F1'],
                horizontalalignment='center',
                rotation=self.xrotation)
        else:
            self.ax1v.set_xticks([0])
            self.ax1v.set_xticklabels(
                ['F1'],
                horizontalalignment='left',
                rotation=self.xrotation)
        self.ax1v.set_yticks([])

        for i in range(self.cm.shape[0]):
            txt = str(int(_nan_to_num(verticalPlot[i])))

            self.ax1v.text(
                0,
                i,
                txt,
                horizontalalignment="center",
                color="white" if verticalPlot[i] > 50 else "black",
                va='center')
        self.axes.append(self.ax1v)

    def add_accuracy(self, thresold=50):
        """
        Add user and producer accuracy.

        Parameters
        ----------
        thresold : int, default 50
            The thresold value where text will be in white instead of black.

        Examples
        --------
        >>> plot.add_accuracy()
        """
        if self.subplot is not False:
            raise Warning(
                'You can\'t add two subplots. You already had ' + str(self.subplot))
        if self.cm_.shape[0] != self.cm_.shape[1]:
            raise Warning('Number of lines and columns must be equal')

        self.subplot = 'F1'

        self.ax1v = plt.subplot(self.gs[0, 1])
        self.ax1h = plt.subplot(self.gs[1, 0])

        self.ax1v.imshow(np.array(np.diag(self.cm_) / np.nansum(self.cm_,
                                                                axis=1) * 100).reshape(-1,
                                                                                       1),
                         cmap=self.diag_color,
                         interpolation='nearest',
                         aspect='equal',
                         vmin=0,
                         vmax=100)

        self.ax1h.imshow(np.array(np.diag(self.cm_) / np.nansum(self.cm_, axis=0) * 100).reshape(
            1, -1), cmap=self.diag_color, interpolation='nearest', aspect='equal', vmin=0, vmax=100)

        self.ax1v.set_yticks(np.arange(self.cm_.shape[0]))
        self.ax1v.set_xticks([])

        self.ax1h.set_yticks([0])
        self.ax1h.set_xticks([])

        for i in range(self.cm.shape[0]):
            iVal = np.int(_nan_to_num(np.array(
                np.diag(self.cm_) / np.nansum(self.cm_, axis=1) * 100).reshape(-1, 1)[i][0], nan=0))

            self.ax1v.text(0, i, iVal, color="white" if iVal >
                           thresold else 'black', ha='center', va='center')

        self.ax1v.set_yticklabels([])
        for j in range(self.cm.shape[1]):
            jVal = np.int(_nan_to_num(np.array(
                np.diag(self.cm_) / np.nansum(self.cm_, axis=0) * 100).reshape(-1, 1)[j][0], nan=0))

            self.ax1h.text(j, 0, jVal, color="white" if jVal >
                           thresold else 'black', ha='center', va='center')

        self.ax1h.set_yticklabels(
            ['Prod\'s acc.'],
            rotation=self.yrotation,
            ha='right',
            va='center')

        self.ax1v.xaxis.set_ticks_position('top')  # THIS IS THE ONLY CHANGE
        self.ax1v.set_xticks([0])
        if self.xrotation < 60:
            ha = 'left'
        else:
            ha = 'center'
        self.ax1v.set_xticklabels(
            ['User\'s acc.'],
            horizontalalignment='left',
            rotation=self.xrotation,
            ha=ha)
        self.axes.append([self.ax1v, self.ax1h])

    def color_diagonal(self, diag_color=plt.cm.Greens,
                       matrix_color=plt.cm.Reds):
        """
        Add user and producer accuracy.

        Parameters
        ----------
        diag_color : pyplot colormap, default plt.cm.Greens.
        matrix_color : pyplot colormap, default plt.cm.Reds

        Examples
        --------
        >>> plot.colorDiag()
        """
        self.diag_color = diag_color
        if self.cm.shape[0] != self.cm.shape[1]:
            raise Warning(
                'Array must have the same number of lines and columns')

        mask = np.zeros(self.cm.shape)
        np.fill_diagonal(mask, 1)

        self.cm2 = np.ma.masked_array(self.cm, mask=np.logical_not(mask))
        self.cm = np.ma.masked_array(self.cm, mask=mask)

        self.ax.imshow(
            self.cm2,
            interpolation='nearest',
            aspect='equal',
            cmap=diag_color,
            vmin=np.amin(
                self.cm_),
            vmax=np.amax(
                self.cm_),
            alpha=1)
        self.ax.imshow(
            self.cm,
            interpolation='nearest',
            aspect='equal',
            cmap=matrix_color,
            vmin=np.amin(
                self.cm_),
            vmax=np.amax(
                self.cm_),
            alpha=1)

    def save_to(self, path, dpi=150):
        """
        Save the plot

        Parameters
        ----------
        path : str
            The path of the file to save.
        dpi : int, default 150.

        Examples
        --------
        >>> plot.saveTo('/tmp/contofu.pdf',dpi=300)
        """
        self.fig.savefig(path, dpi=dpi, bbox_inches='tight')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# =============================================================================
# ___  ___                       _____           _______
# |  \/  |                      |_   _|         | | ___ \
# | .  . |_   _ ___  ___  ___     | | ___   ___ | | |_/ / _____  __
# | |\/| | | | / __|/ _ \/ _ \    | |/ _ \ / _ \| | ___ \/ _ \ \/ /
# | |  | | |_| \__ \  __/ (_) |   | | (_) | (_) | | |_/ / (_) >  <
# \_|  |_/\__,_|___/\___|\___/    \_/\___/ \___/|_\____/ \___/_/\_\
#
# @author:  Nicolas Karasiak
# @site:    www.karasiak.net
# @git:     www.github.com/nkarasiak/MuseoToolBox
# =============================================================================


def push_feedback(msg, feedback=None):
    # in order to convert in Qgis Processing
    # =============================================================================
    #     if feedback and feedback is not True:
    #         if feedback == 'gui':
    #             QgsMessageLog.logMessage(str(msg))
    #         else:
    #             feedback.setProgressText(msg)
    #     else:
    # =============================================================================
    print(msg)


class ProgressBar:
    def __init__(self, total, message='', length=40):
        """
        total : int
            Total number of samples.
        message : str
            Custom message to show before the progress bar.
        length : int.
            Length of the bar.
        """
        self.start = 0
        self.total = total
        self.length = length
        self.message = message
        self.lastPosition = None

    def add_position(self, value=False):
        """
        Add progress to the bar.

        Parameters
        ----------

        value : int or False.
            If false, will add one.
        """

        if value is False:
            self.start += 1
            value = self.start
            inPercent = int(self.start / self.total * 100)

        else:
            inPercent = int(value / self.total * 100)
        if inPercent != self.lastPosition:
            self.lastPosition = inPercent
            self.nHash = int(self.length * (value / self.total))
            self.nPoints = int(self.length - int(self.nHash))

            self.printBar(inPercent)

    def printBar(self, value):
        if value == 100:
            end = "\n"
        else:
            end = "\r"

        # print(self.nHash)
        # print(self.nPoints)
        print('\r' + self.message + ' [{}{}]{}%'.format(self.nHash * "#", self.nPoints * ".", self.lastPosition), end=end, flush=True)



    #!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Nov 24 17:38:05 2019

@author: nicolas

in_image : string.
    A filename or path of a raster file.
    It could be any file that GDAL can open.

in_vector : string.
    A filename or path corresponding to a vector file.
    It could be any file that GDAL/OGR can open.
    
out_image : string.
    A geotiff extension filename corresponding to a raster image to create.
    
X : array-like, shape = [n_samples, n_features]
    The training input samples.

y : array-like, shape = [n_samples]
    The target values.
    
in_image_mask : str
    A filename or path corresponding to a raster image.
    0 values are considered as masked data.

out_image : str
    A filename or path corresponding to a geotiff (.tif) raster image to save.
    0 values are considered as masked data.


"""


